import pyarrow as pa
from typing import List, Optional, Tuple
from deltacat.tests.compute.test_util_common import (
    offer_iso8601_timestamp_list,
    PartitionKey,
    PartitionKeyType,
    assert_compaction_audit,
    assert_compaction_audit_no_hash_bucket,
)
from deltacat.tests.compute.test_util_constant import (
    DEFAULT_MAX_RECORDS_PER_FILE,
    DEFAULT_HASH_BUCKET_COUNT,
)
import string
from dataclasses import dataclass


from deltacat.storage import (
    DeltaType,
    EntryParams,
)
from deltacat.types.media import ContentType

from deltacat.compute.compactor.model.compactor_version import CompactorVersion

from deltacat.storage.model.sort_key import SortKey

from deltacat.utils.pyarrow import (
    ReadKwargsProviderPyArrowSchemaOverride,
    content_type_to_reader_kwargs,
    pyarrow_read_csv,
)
from deltacat.tests.compute.compact_partition_test_cases import (
    BaseCompactorTestCase,
    with_compactor_version_func_test_param,
    ZERO_VALUED_SORT_KEY,
    ZERO_VALUED_PARTITION_VALUES_PARAM,
    ZERO_VALUED_PARTITION_KEYS_PARAM,
    ZERO_VALUED_PRIMARY_KEY,
    EMPTY_UTSV_PATH,
)
from deltacat.exceptions import ValidationError


@dataclass(frozen=True)
class RebaseThenIncrementalCompactionTestCaseParams(BaseCompactorTestCase):
    """
    A pytest parameterized test case for the `compact_partition` function with rebase and incremental compaction.

    Args:
        * (inherited from CompactorTestCase): see CompactorTestCase docstring for details
        incremental_deltas: List[Tuple[pa.Table, DeltaType, Optional[Dict[str, str]]]] - argument required for delta creation during the incremental phase of compact_partition test setup. Incoming deltas during incremental expressed as a pyarrow array
        rebase_expected_compact_partition_result: pa.Table - expected table after rebase compaction runs. An output that is asserted on in Rebase then Incremental unit tests
    """

    incremental_deltas: List[Tuple[pa.Table, DeltaType, Optional[EntryParams]]]
    rebase_expected_compact_partition_result: pa.Table


REBASE_THEN_INCREMENTAL_TEST_CASES = {
    "1-rebase-then-incremental-sanity": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key=["sk_col_1"]),
            SortKey.of(key=["sk_col_2"]),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([str(i) for i in range(10)]),
                        pa.array([i for i in range(20, 30)]),
                        pa.array(["foo"] * 10),
                        pa.array([i / 10 for i in range(40, 50)]),
                    ],
                    names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            )
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(20, 30)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(40, 50)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit,
    ),
    "2-rebase-then-incremental-pk-multi": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1", "pk_col_2"},
        sort_keys=[
            SortKey.of(key=["sk_col_1"]),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i % 4) for i in range(10)]),
                pa.array([(i % 4) / 10 for i in range(9, -1, -1)]),
                pa.array(offer_iso8601_timestamp_list(10, "minutes")),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "pk_col_2", "sk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array(["0", "1", "2", "3"]),
                pa.array([0.1, 0, 0.3, 0.2]),
                pa.array(
                    [
                        "2023-05-03T10:00:00Z",
                        "2023-05-03T09:59:00Z",
                        "2023-05-03T09:58:00Z",
                        "2023-05-03T09:57:00Z",
                    ]
                ),
                pa.array([1, 1.1, 1.2, 1.3]),
            ],
            names=["pk_col_1", "pk_col_2", "sk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array(["0", "1", "2", "3"]),
                        pa.array([0.1, 0, 0.3, 0.2]),
                        pa.array(
                            [
                                "2023-05-03T10:00:00Z",
                                "2023-05-03T09:59:00Z",
                                "2023-05-03T09:58:00Z",
                                "2023-05-03T09:57:00Z",
                            ]
                        ),
                        pa.array([1, 1.1, 1.2, 1.3]),
                    ],
                    names=["pk_col_1", "pk_col_2", "sk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                {},
            )
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array(["0", "1", "2", "3"]),
                pa.array([0.1, 0, 0.3, 0.2]),
                pa.array(
                    [
                        "2023-05-03T10:00:00Z",
                        "2023-05-03T09:59:00Z",
                        "2023-05-03T09:58:00Z",
                        "2023-05-03T09:57:00Z",
                    ]
                ),
                pa.array([1, 1.1, 1.2, 1.3]),
            ],
            names=["pk_col_1", "pk_col_2", "sk_col_1", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit,
    ),
    "3-rebase-then-incremental-no-sk-no-partition-key": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=ZERO_VALUED_PARTITION_KEYS_PARAM,
        partition_values=ZERO_VALUED_PARTITION_VALUES_PARAM,
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i % 4) for i in range(12)]),
                pa.array([i / 10 for i in range(10, 22)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array(["0", "1", "2", "3"]),
                pa.array([1.8, 1.9, 2.0, 2.1]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array(["0", "1", "2", "3"]),
                        pa.array([18.0, 19.0, 20.0, 21.0]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            )
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array(["0", "1", "2", "3"]),
                pa.array([18.0, 19.0, 20.0, 21.0]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit,
    ),
    "4-rebase-then-incremental-partial-deltas-on-incremental-deltas": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i / 10 for i in range(10)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i / 10 for i in range(10)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array(["8", "9"]),
                        pa.array([200.0, 100.0]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            )
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i / 10 for i in range(8)] + [200.0] + [100.0]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit,
    ),
    "5-rebase-then-incremental-partial-deltas-on-incremental-deltas-2": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key=["sk_col_1"]),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([i % 4 for i in range(12)]),
                pa.array([(i / 10 * 10) % 4 for i in range(12)][::-1]),
                pa.array(list(string.ascii_lowercase)[:12]),
            ],
            names=["pk_col_1", "sk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3]),
                pa.array([3.0, 2.0, 1.0, 0.0]),
                pa.array(["i", "j", "k", "l"]),
            ],
            names=["pk_col_1", "sk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([1, 4]),
                        pa.array([4.0, 2.0]),
                        pa.array(["a", "b"]),
                    ],
                    names=["pk_col_1", "sk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            )
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3, 4]),
                pa.array([3.0, 4.0, 1.0, 0.0, 2.0]),
                pa.array(["i", "a", "k", "l", "b"]),
            ],
            names=["pk_col_1", "sk_col_1", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit,
    ),
    "6-rebase-then-incremental-hash-bucket-GT-records-per-compacted-file-v2-only": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key=["sk_col_1"]),
            SortKey.of(key=["sk_col_2"]),
        ],
        partition_keys=[PartitionKey.of("day", PartitionKeyType.TIMESTAMP)],
        partition_values=["2022-01-01T00:00:00.000Z"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(12)]),
                pa.array([i for i in range(0, 12)]),
                pa.array(["foo"] * 12),
                pa.array([i / 10 for i in range(10, 22)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(12)]),
                pa.array([i for i in range(0, 12)]),
                pa.array(["foo"] * 12),
                pa.array([i / 10 for i in range(10, 22)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([str(i) for i in range(12)]),
                        pa.array([i for i in range(20, 32)]),
                        pa.array(["foo"] * 12),
                        pa.array([i / 10 for i in range(40, 52)]),
                    ],
                    names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            )
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(12)]),
                pa.array([i for i in range(20, 32)]),
                pa.array(["foo"] * 12),
                pa.array([i / 10 for i in range(40, 52)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=10,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT + 10,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit,
    ),
    "7-rebase-then-incremental-no-pk-compactor-v2-only": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys=ZERO_VALUED_PRIMARY_KEY,
        sort_keys=[
            SortKey.of(key=["sk_col_1"]),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([1, 2, 3]),
                pa.array([1.0, 2.0, 3.0]),
            ],
            names=["sk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([1, 1, 2, 2, 3, 3]),
                pa.array([1.0, 1.0, 2.0, 2.0, 3.0, 3.0]),
            ],
            names=["sk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([4, 5, 6]),
                        pa.array([10.0, 11.0, 12.0]),
                    ],
                    names=["sk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            )
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([1, 1, 2, 2, 3, 3, 4, 5, 6]),
                pa.array([1.0, 1.0, 2.0, 2.0, 3.0, 3.0, 10.0, 11.0, 12.0]),
            ],
            names=["sk_col_1", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=10,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit,
    ),
    "8-rebase-then-incremental-empty-csv-delta-case": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pyarrow_read_csv(
                    EMPTY_UTSV_PATH,
                    **ReadKwargsProviderPyArrowSchemaOverride(
                        schema=pa.schema(
                            [
                                ("pk_col_1", pa.string()),
                                ("col_1", pa.float64()),
                            ]
                        )
                    )(
                        ContentType.UNESCAPED_TSV.value,
                        content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value),
                    ),
                ),
                DeltaType.UPSERT,
                None,
            )
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array(["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]),
                pa.array([1.0, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=None,
    ),
    "9-rebase-then-incremental-single-hash-bucket": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key=["sk_col_1"]),
            SortKey.of(key=["sk_col_2"]),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([str(i) for i in range(10)]),
                        pa.array([i for i in range(20, 30)]),
                        pa.array(["foo"] * 10),
                        pa.array([i / 10 for i in range(40, 50)]),
                    ],
                    names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            )
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(20, 30)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(40, 50)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=1,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=None,
    ),
    "10-rebase-then-incremental-drop-duplicates-false-on-incremental-v2-only": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key=["sk_col_1"]),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([(i % 4) for i in range(8)]),
                pa.array([(i % 2) for i in range(8)]),
                pa.array([i / 10 for i in range(10, 18)]),
            ],
            names=["pk_col_1", "sk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3]),
                pa.array([0, 1, 0, 1]),
                pa.array([1.4, 1.5, 1.6, 1.7]),
            ],
            names=["pk_col_1", "sk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([0, 1, 2, 3, 1]),
                        pa.array([0, 1, 0, 1, 0]),
                        pa.array([i / 10 for i in range(20, 25)]),
                    ],
                    names=["pk_col_1", "sk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            )
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 0, 1, 1, 1, 2, 2, 3, 3]),
                pa.array([0, 0, 1, 0, 1, 0, 0, 1, 1]),
                pa.array([1.4, 2, 1.5, 2.4, 2.1, 1.6, 2.2, 1.7, 2.3]),
            ],
            names=["pk_col_1", "sk_col_1", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=False,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit,
    ),
    "11-rebase-then-empty-incremental-delta": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key=["sk_col_1"]),
            SortKey.of(key=["sk_col_2"]),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        incremental_deltas=[(None, DeltaType.UPSERT, None)],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=3,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit_no_hash_bucket,
    ),
    "12-rebase-then-incremental-hash-bucket-single": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key=["sk_col_1"]),
            SortKey.of(key=["sk_col_2"]),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([str(i) for i in range(10)]),
                        pa.array([i for i in range(20, 30)]),
                        pa.array(["foo"] * 10),
                        pa.array([i / 10 for i in range(40, 50)]),
                    ],
                    names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            )
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(20, 30)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(40, 50)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=1,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=None,
    ),
    "13-rebase-then-empty-incremental-delta-hash-bucket-single": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key=["sk_col_1"]),
            SortKey.of(key=["sk_col_2"]),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        incremental_deltas=[(None, DeltaType.UPSERT, None)],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=1,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=None,
    ),
    "14-rebase-then-incremental-with-null-pk": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key=["sk_col_1"]),
            SortKey.of(key=["sk_col_2"]),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(9)] + [None]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(9)] + [None]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([str(i) for i in range(9)] + [None]),
                        pa.array([i for i in range(20, 30)]),
                        pa.array(["foo"] * 10),
                        pa.array([i / 10 for i in range(40, 50)]),
                    ],
                    names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            )
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(9)] + [None]),
                pa.array([i for i in range(20, 30)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(40, 50)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit,
    ),
}

REBASE_THEN_INCREMENTAL_DELETE_DELTA_TYPE_TEST_CASES = {
    "14-rebase-then-incremental-delete-type-delta-on-incremental": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([i for i in range(12)]),
                pa.array([str(i) for i in range(0, 12)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([i for i in range(12)]),
                pa.array([str(i) for i in range(0, 12)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([10, 11, 12, 13]),
                        pa.array(["a", "b", "c", "d"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
            (
                pa.Table.from_arrays(
                    [pa.array([10, 11]), pa.array(["a", "b"])],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["pk_col_1", "col_1"]),
            ),
            (
                pa.Table.from_arrays(
                    [pa.array(["c"])],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
            (
                pa.Table.from_arrays(
                    [pa.array(["c"])],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
            (
                pa.Table.from_arrays(
                    [pa.array(["c"])],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
            (
                pa.Table.from_arrays(
                    [pa.array([10, 11]), pa.array(["a", "b"])],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["pk_col_1", "col_1"]),
            ),
            (
                pa.Table.from_arrays(
                    [pa.array(["c"])],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([i for i in range(10)] + [13]),
                pa.array([str(i) for i in range(0, 10)] + ["d"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit,
    ),
    "15-rebase-then-incremental-delete-type-delta-on-incremental-multi-pk": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1", "pk_col_2"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.TIMESTAMP)],
        partition_values=["2022-01-01T00:00:00.000Z"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([(i % 4) for i in range(12)]),
                pa.array([float(i % 4) for i in range(12, 0, -1)]),
                pa.array([str(i) for i in range(0, 12)]),
            ],
            names=["pk_col_1", "pk_col_2", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3]),
                pa.array([0.0, 3.0, 2.0, 1.0]),
                pa.array(["8", "9", "10", "11"]),
            ],
            names=["pk_col_1", "pk_col_2", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [  # delete last two primary keys
                        pa.array(["10", "11"]),
                    ],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            )
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1]),
                pa.array([0.0, 3.0]),
                pa.array(["8", "9"]),
            ],
            names=["pk_col_1", "pk_col_2", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit_no_hash_bucket,
    ),
    "16-rebase-then-incremental-delete-type-delta-on-incremental-multi-pk-delete-all": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1", "pk_col_2"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.TIMESTAMP)],
        partition_values=["2022-01-01T00:00:00.000Z"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([(i % 4) for i in range(12)]),
                pa.array([float(i % 4) for i in range(12, 0, -1)]),
                pa.array([str(i) for i in range(0, 12)]),
            ],
            names=["pk_col_1", "pk_col_2", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3]),
                pa.array([0.0, 3.0, 2.0, 1.0]),
                pa.array(["8", "9", "10", "11"]),
            ],
            names=["pk_col_1", "pk_col_2", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array(["8", "9", "10", "11"]),
                    ],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            )
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([]),
                pa.array([]),
                pa.array([]),
            ],
            schema=pa.schema(
                [
                    ("pk_col_1", pa.int64()),
                    ("pk_col_2", pa.float64()),
                    ("col_1", pa.string()),
                ]
            ),
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit_no_hash_bucket,
    ),
    "17-rebase-then-incremental-delete-type-delta-delete-entire-base-table": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.TIMESTAMP)],
        partition_values=["2022-01-01T00:00:00.000Z"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([(i % 4) for i in range(1000)] + [4, 5]),
                pa.array([str(i) for i in range(0, 1000)] + ["fiz", "buz"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3, 4, 5]),
                pa.array(["996", "997", "998", "999", "fiz", "buz"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array(["996", "997", "998", "999", "fiz", "buz"]),
                    ],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            )
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([]),
                pa.array([]),
            ],
            schema=pa.schema(
                [
                    ("pk_col_1", pa.int64()),
                    ("col_1", pa.string()),
                ]
            ),
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=True,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit_no_hash_bucket,
    ),
    "18-rebase-then-incremental-delete-type-delta-keep-base-table-drop-all-incremental": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.TIMESTAMP)],
        partition_values=["2022-01-01T00:00:00.000Z"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([(i % 4) for i in range(1000)] + [4, 5]),
                pa.array([str(i) for i in range(0, 1000)] + ["fiz", "buz"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3, 4, 5]),
                pa.array(["996", "997", "998", "999", "fiz", "buz"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([0, 1, 2]),
                        pa.array(["0", "1", "2"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array([2, 3]),
                        pa.array(["abc", "def"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array([4]),
                        pa.array(["ghi"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array(["0", "1", "2", "abc", "def", "ghi"]),
                    ],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([5]),
                pa.array(["buz"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=True,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit,
    ),
    "19-rebase-then-incremental-delete-type-delta-drop-only-from-base-table-keep-all-incremental": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.TIMESTAMP)],
        partition_values=["2022-01-01T00:00:00.000Z"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([(i % 4) for i in range(1000)] + [4, 5]),
                pa.array([str(i) for i in range(0, 1000)] + ["fiz", "buz"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3, 4, 5]),
                pa.array(["996", "997", "998", "999", "fiz", "buz"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([0, 1, 2, 3, 4, 5]),
                        pa.array(["0", "1", "2", "3", "4", "5"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array([0]),
                        pa.array(["foo"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array(["996", "997", "998", "999", "fiz", "buz"]),
                    ],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3, 4, 5]),
                pa.array(["foo", "1", "2", "3", "4", "5"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=True,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit,
    ),
    "20-rebase-then-incremental-delete-type-delta-drop-all-base-table-drop-all-incremental": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.TIMESTAMP)],
        partition_values=["2022-01-01T00:00:00.000Z"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([(i % 4) for i in range(1000)] + [4, 5]),
                pa.array([str(i) for i in range(0, 1000)] + ["fiz", "buz"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3, 4, 5]),
                pa.array(["996", "997", "998", "999", "fiz", "buz"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([6]),
                        pa.array(["foo"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array(["996", "997", "998", "999", "fiz", "buz", "foo"]),
                    ],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([]),
                pa.array([]),
            ],
            schema=pa.schema(
                [
                    ("pk_col_1", pa.int64()),
                    ("col_1", pa.string()),
                ]
            ),
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=True,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit,
    ),
    "21-rebase-then-incremental-delete-type-delta-UDDUUDD": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.TIMESTAMP)],
        partition_values=["2022-01-01T00:00:00.000Z"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([(i % 4) for i in range(1000)] + [4, 5]),
                pa.array([str(i) for i in range(0, 1000)] + ["fiz", "buz"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3, 4, 5]),
                pa.array(["996", "997", "998", "999", "fiz", "buz"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([0, 1, 2, 3, 4, 5]),
                        pa.array(["0", "1", "2", "3", "4", "5"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array(["DOESNOTEXIST"]),
                    ],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array([1]),
                        pa.array(["1"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array([2, 3, 6, 7]),
                        pa.array(["boo", "bar", "fiz", "aaa"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array(["fiz", "bar", "boo"]),
                    ],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 4, 5, 7]),
                pa.array(["0", "4", "5", "aaa"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=True,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit,
    ),
    "22-rebase-then-incremental-delete-type-delta-UD-affects-compacted-and-incremental": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.TIMESTAMP)],
        partition_values=["2022-01-01T00:00:00.000Z"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([0, 1]),
                pa.array(["0", "1"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1]),
                pa.array(["0", "1"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([0]),
                        pa.array(["1"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array(["1"]),
                    ],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([]),
                pa.array([]),
            ],
            schema=pa.schema(
                [
                    ("pk_col_1", pa.int64()),
                    ("col_1", pa.string()),
                ]
            ),
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=True,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit,
    ),
    "23-rebase-then-incremental-delete-type-delta-UDU-upsert-again": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.TIMESTAMP)],
        partition_values=["2022-01-01T00:00:00.000Z"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([0, 1]),
                pa.array(["0", "1"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1]),
                pa.array(["0", "1"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([0]),
                        pa.array(["1"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array(["1"]),
                    ],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array([0]),
                        pa.array(["1"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0]),
                pa.array(["1"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=True,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit,
    ),
    "24-rebase-then-incremental-delete-type-no-delete-column-has-delete-deltas-expected-exception": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.TIMESTAMP)],
        partition_values=["2022-01-01T00:00:00.000Z"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([0, 1]),
                pa.array(["0", "1"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1]),
                pa.array(["0", "1"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([0]),
                        pa.array(["1"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array(["1"]),
                    ],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                None,
            ),
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([]),
                pa.array([]),
            ],
            schema=pa.schema(
                [
                    ("pk_col_1", pa.int64()),
                    ("col_1", pa.string()),
                ]
            ),
        ),
        expected_terminal_exception=ValidationError,
        expected_terminal_exception_message="One of the assertions in DeltaCAT has failed",
        do_create_placement_group=True,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit,
    ),
    "25-rebase-then-incremental-delete-type-delta-has-delete-column-no-delete-records": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.TIMESTAMP)],
        partition_values=["2022-01-01T00:00:00.000Z"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2]),
                pa.array(["0", "1", "2"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2]),
                pa.array(["0", "1", "2"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([0]),
                        pa.array(["1"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array([]),
                        pa.array([]),
                    ],
                    schema=pa.schema(
                        [
                            ("pk_col_1", pa.int64()),
                            ("col_1", pa.string()),
                        ]
                    ),
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2]),
                pa.array(["1", "1", "2"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=True,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit,
    ),
    "26-rebase-then-incremental-delete-type-delta-UDU-duplicate-delete-records": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.TIMESTAMP)],
        partition_values=["2022-01-01T00:00:00.000Z"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3]),
                pa.array(["0", "1", "2", "3"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3]),
                pa.array(["0", "1", "2", "3"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([0]),
                        pa.array(["1"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array(["1", "1", "1", "2", "2", "2"]),
                    ],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([3]),
                pa.array(["3"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=True,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit,
    ),
    "27-rebase-then-incremental-delete-type-delta-DDU-deletes-then-upserts": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.TIMESTAMP)],
        partition_values=["2022-01-01T00:00:00.000Z"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3, 4]),
                pa.array(["0", "1", "2", "3", "4"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3, 4]),
                pa.array(["0", "1", "2", "3", "4"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array(["1", "1", "2"]),
                    ],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array([0]),
                    ],
                    names=["pk_col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["pk_col_1"]),
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array([0, 3]),
                        pa.array(["a", "b"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 3, 4]),
                pa.array(["a", "b", "4"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=True,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit,
    ),
    "28-rebase-then-incremental-delete-type-delta-hash-bucket-single": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key=["sk_col_1"]),
            SortKey.of(key=["sk_col_2"]),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([str(i) for i in range(10)]),
                        pa.array([i for i in range(20, 30)]),
                        pa.array(["foo"] * 10),
                        pa.array([i / 10 for i in range(40, 50)]),
                    ],
                    names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array([i / 10 for i in range(0, 3)]),
                    ],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array([i / 10 for i in range(43, 45)]),
                    ],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array([i for i in range(20, 25)]),
                    ],
                    names=["sk_col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["sk_col_1"]),
            ),
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(5, 10)]),
                pa.array([i for i in range(25, 30)]),
                pa.array(["foo"] * 5),
                pa.array([i / 10 for i in range(45, 50)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=1,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=None,
    ),
    "29-rebase-then-incremental-delete-type-delta-no-pk-compactor": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys=ZERO_VALUED_PRIMARY_KEY,
        sort_keys=[
            SortKey.of(key=["sk_col_1"]),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([1, 2, 3]),
                pa.array([1.0, 2.0, 3.0]),
            ],
            names=["sk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([1, 1, 2, 2, 3, 3]),
                pa.array([1.0, 1.0, 2.0, 2.0, 3.0, 3.0]),
            ],
            names=["sk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([3.0]),
                    ],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array([4, 5, 6]),
                        pa.array([10.0, 11.0, 12.0]),
                    ],
                    names=["sk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
            (
                pa.Table.from_arrays(
                    [
                        pa.array([6]),
                    ],
                    names=["sk_col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["sk_col_1"]),
            ),
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([1, 1, 2, 2, 4, 5]),
                pa.array([1.0, 1.0, 2.0, 2.0, 10.0, 11.0]),
            ],
            names=["sk_col_1", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=10,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit,
    ),
    "30-rebase-then-incremental-delete-type-delta-on-incremental-compactor-v1-v2": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([i for i in range(12)]),
                pa.array([str(i) for i in range(0, 12)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([i for i in range(12)]),
                pa.array([str(i) for i in range(0, 12)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([10, 11]),
                        pa.array(["10", "11"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["pk_col_1", "col_1"]),
            ),
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([i for i in range(10)]),
                pa.array([str(i) for i in range(0, 10)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit_no_hash_bucket,
    ),
    "31-rebase-then-incremental-delete-delta-on-incremental-null-pk-delete-null": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([i for i in range(11)] + [None]),
                pa.array([str(i) for i in range(0, 12)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([i for i in range(11)] + [None]),
                pa.array([str(i) for i in range(0, 12)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([10, 11, None, 13]),
                        pa.array(["a", "b", "c", "d"]),
                    ],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.UPSERT,
                None,
            ),
            (
                pa.Table.from_arrays(
                    [pa.array([10, 11]), pa.array(["a", "b"])],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["pk_col_1", "col_1"]),
            ),
            (
                pa.Table.from_arrays(
                    [pa.array([None])],  # Support deleting null PK records
                    names=["pk_col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["pk_col_1"]),
            ),
            (
                pa.Table.from_arrays(
                    [pa.array(["c"])],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
            (
                pa.Table.from_arrays(
                    [pa.array(["c"])],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
            (
                pa.Table.from_arrays(
                    [pa.array([10, 11]), pa.array(["a", "b"])],
                    names=["pk_col_1", "col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["pk_col_1", "col_1"]),
            ),
            (
                pa.Table.from_arrays(
                    [pa.array(["c"])],
                    names=["col_1"],
                ),
                DeltaType.DELETE,
                EntryParams.of(["col_1"]),
            ),
        ],
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([i for i in range(10)] + [13]),
                pa.array([str(i) for i in range(0, 10)] + ["d"]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit,
    ),
}

REBASE_THEN_INCREMENTAL_TEST_CASES = with_compactor_version_func_test_param(
    {
        **REBASE_THEN_INCREMENTAL_TEST_CASES,
        **REBASE_THEN_INCREMENTAL_DELETE_DELTA_TYPE_TEST_CASES,
    },
)



from typing import List

import pyarrow as pa
from pyarrow import feather

from deltacat.experimental.storage.rivulet.metastore.sst import SSTableRow
from deltacat.experimental.storage.rivulet import Schema
from deltacat.experimental.storage.rivulet.arrow.serializer import ArrowSerializer
from deltacat.experimental.storage.rivulet.fs.file_provider import FileProvider


class FeatherDataSerializer(ArrowSerializer):
    """
    Feather data writer. Responsible for flushing rows to feather files and returning SSTable rows for any file(s) written

    TODO Support recording byte range offsets. Deferring this for now
        We may need to provide a wrapper class over fsspec which introspects how many bytes written
        when .write is called on output stream
    """

    def __init__(self, file_provider: FileProvider, schema: Schema):
        super().__init__(file_provider, schema)

    def serialize(self, table: pa.Table) -> List[SSTableRow]:
        file = self.file_provider.provide_data_file("feather")

        with file.create() as outfile:
            # Note that write_feather says that dest is a string, but it is really any object implementing write()
            feather.write_feather(table, dest=outfile)

        # Because we only write one row group, it only creates one SSTableRow
        #  we may have more granular SST indexes for other formats
        key_min, key_max = self._get_min_max_key(table)
        # TODO need to populate byte offsets. For now, we are writing single files per SSTableRow
        return [SSTableRow(key_min, key_max, file.location, 0, 0)]



"""
Common utility functions for the Iceberg converter example.
"""

import random
import string
import time
from pyiceberg.catalog import load_catalog
import requests
from deltacat import local_job_client
from deltacat.constants import DEFAULT_NAMESPACE
from deltacat.experimental.converter_agent.table_monitor import _generate_job_name


def generate_random_suffix(length=8):
    """Generate a random string of specified length using letters and digits."""
    return "".join(random.choices(string.ascii_lowercase + string.digits, k=length))


def check_rest_catalog():
    """Check if REST catalog is running."""
    try:
        response = requests.get("http://localhost:8181/v1/config", timeout=5)
        if response.status_code == 200:
            print("✅ REST catalog is running")
            return True
    except requests.exceptions.RequestException:
        pass

    print("❌ REST catalog is not running")
    print(
        "📋 Start it with: docker run -d -p 8181:8181 --name iceberg-rest-catalog tabulario/iceberg-rest:1.6.0"
    )
    return False


def wait_for_deltacat_jobs(
    table_name, warehouse_path="/tmp/iceberg_rest_warehouse", timeout=120
):
    """
    Wait for DeltaCAT converter jobs to complete by checking job status.

    Args:
        table_name: Full table name (e.g., "default.demo_table_abc123")
        warehouse_path: Warehouse path used for job tracking
        timeout: Maximum seconds to wait for job completion

    Returns:
        True if all jobs completed, False if timeout
    """
    print(f"\n⏳ Monitoring DeltaCAT converter jobs for table: {table_name}")

    # Parse table name to get namespace and table name
    if "." in table_name:
        namespace, actual_table_name = table_name.split(".", 1)
    else:
        namespace = DEFAULT_NAMESPACE
        actual_table_name = table_name

    # Create job key matching the format used in managed.py
    job_name = _generate_job_name(
        warehouse_path=warehouse_path, namespace=namespace, table_name=actual_table_name
    )

    start_time = time.time()

    try:
        # Get the job client
        client = local_job_client(ray_init_args={"local_mode": True})

        while time.time() - start_time < timeout:
            job_details_list = client.list_jobs()
            print(f"🔍 Job details list: {job_details_list}")
            job_submission_ids = [
                job_details.submission_id for job_details in job_details_list
            ]

            # Check if we have any tracked jobs for this table
            print(f"🔍 Looking for submission ID: {job_name} in {job_submission_ids}")
            if job_name in job_submission_ids:
                # Check job status with Ray
                try:
                    job_status = client.get_job_status(job_name)
                    print(f"📊 Job {job_name} status: {job_status}")
                    # Check if job is still running
                    if job_status and str(job_status) in ["PENDING", "RUNNING"]:
                        time.sleep(2)  # Short polling interval
                        continue
                    else:
                        print(f"✅ Job {job_name} completed with status: {job_status}")
                        return True

                except Exception as e:
                    print(f"⚠️  Could not check job status for {job_name}: {e}")
                    # If we can't check status, assume job is done
                    return True
            time.sleep(1)
        print(f"⏰ Timeout waiting for DeltaCAT job completion after {timeout} seconds")
        return False

    except Exception as e:
        print(f"❌ Error monitoring DeltaCAT jobs: {e}")
        # Fall back to short sleep if monitoring fails
        print(f"🔄 Falling back to {timeout}-second wait...")
        time.sleep(timeout)
        return True


def verify_duplicate_resolution(
    table_name, warehouse_path="/tmp/iceberg_rest_warehouse"
):
    """
    Verify that the DeltaCAT converter successfully resolved duplicates.
    """
    try:
        print(f"\n🔍 Verifying duplicate resolution for table: {table_name}")

        # Create PyIceberg catalog to check results
        verification_catalog = load_catalog(
            "workflow_verification_catalog",
            **{
                "type": "rest",
                "warehouse": warehouse_path,
                "uri": "http://localhost:8181",
            },
        )

        # Load the table and scan its contents
        table_identifier = table_name
        tbl = verification_catalog.load_table(table_identifier)
        scan_result = tbl.scan().to_arrow().to_pydict()

        # Check the results
        result_ids = sorted(scan_result["id"])
        unique_ids = sorted(set(result_ids))

        print(f"📊 Final verification results:")
        print(f"  - Total records: {len(result_ids)}")
        print(f"  - Unique IDs: {len(unique_ids)}")
        print(f"  - IDs found: {result_ids}")

        # Check if duplicates were resolved
        expected_unique_ids = [1, 2, 3, 4, 5, 6, 7, 8, 9]
        if result_ids == expected_unique_ids:
            # Verify that the latest versions were preserved
            names_by_id = {}
            versions_by_id = {}
            for i, id_val in enumerate(scan_result["id"]):
                names_by_id[id_val] = scan_result["name"][i]
                versions_by_id[id_val] = scan_result["version"][i]

            if (
                names_by_id.get(2) == "Robert"
                and names_by_id.get(3) == "Charles"
                and versions_by_id.get(2) == 2
                and versions_by_id.get(3) == 2
            ):
                print(f"✅ Duplicate resolution SUCCESSFUL!")
                print(f"  - All 9 IDs are unique")
                print(f"  - Latest versions preserved (Bob→Robert, Charlie→Charles)")
                print(f"  - Version numbers correct (v2 for updated records)")
                return True
            else:
                print(f"❌ Latest versions not preserved correctly")
        else:
            print(f"❌ Duplicates still present or unexpected record count")
            print(f"  - Expected: {expected_unique_ids}")
            print(f"  - Got: {result_ids}")

        return False

    except Exception as e:
        print(f"❌ Error during verification: {e}")
        return False



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Optional, Any, List, Tuple, Dict

from pyarrow.compute import SortOptions

from deltacat.storage.model.types import (
    SortOrder,
    NullOrder,
)
from deltacat.storage.model.schema import FieldLocator
from deltacat.storage.model.transform import Transform

UNSORTED_SCHEME_NAME = "unsorted_scheme"
UNSORTED_SCHEME_ID = "deadbeef-7277-49a4-a195-fdc8ed235d42"


class SortKey(tuple):
    @staticmethod
    def of(
        key: Optional[List[FieldLocator]],
        sort_order: SortOrder = SortOrder.ASCENDING,
        null_order: NullOrder = NullOrder.AT_END,
        transform: Optional[Transform] = None,
        native_object: Optional[Any] = None,
    ) -> SortKey:
        """
        Create a sort key from a field name to use as the sort key, and
        the sort order for this key. If no sort order is specified, then the
        data will be sorted in ascending order by default.
        """
        return SortKey(
            (
                key,
                sort_order.value if isinstance(sort_order, SortOrder) else sort_order,
                null_order.value if isinstance(null_order, NullOrder) else null_order,
                transform,
                native_object,
            )
        )

    def equivalent_to(
        self,
        other: SortKey,
    ):
        if other is None:
            return False
        if not isinstance(other, tuple):
            return False
        if not isinstance(other, SortKey):
            other = SortKey(other)
        return (
            self.key == other.key
            and self.transform == other.transform
            and self.sort_order == other.sort_order
            and self.null_order == other.null_order
        )

    @property
    def key(self) -> Optional[List[FieldLocator]]:
        return self[0]

    @property
    def sort_order(self) -> SortOrder:
        return SortOrder(self[1])

    @property
    def null_order(self) -> NullOrder:
        return NullOrder(self[2])

    @property
    def transform(self) -> Optional[Transform]:
        val: Dict[str, Any] = (
            Transform(self[3]) if len(self) >= 4 and self[3] is not None else None
        )
        return val

    @property
    def arrow(self) -> List[Tuple[str, str]]:
        # TODO(pdames): Convert unsupported field locators to arrow field names,
        #   and transforms/multi-key-sorts to pyarrow compute expressions. Add
        #   null order via SortOptions when supported per field by Arrow.
        return (
            [(field_locator, self[1]) for field_locator in self[0]] if self[0] else []
        )

    @property
    def native_object(self) -> Optional[Any]:
        return self[4] if len(self) >= 5 else None


class SortKeyList(List[SortKey]):
    @staticmethod
    def of(items: List[SortKey]) -> SortKeyList:
        typed_items = SortKeyList()
        for item in items:
            if item is not None and not isinstance(item, SortKey):
                item = SortKey(item)
            typed_items.append(item)
        return typed_items

    def __getitem__(self, item):
        val = super().__getitem__(item)
        if val is not None and not isinstance(val, SortKey):
            self[item] = val = SortKey(val)
        return val

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]  # This triggers __getitem__ conversion


class SortScheme(dict):
    @staticmethod
    def of(
        keys: Optional[SortKeyList],
        name: Optional[str] = None,
        scheme_id: Optional[str] = None,
        native_object: Optional[Any] = None,
    ) -> SortScheme:
        # Validate keys if provided
        if keys is not None:
            # Check for empty keys list
            if len(keys) == 0:
                raise ValueError("Sort scheme cannot have empty keys list")

            # Check for duplicate keys
            key_names = []
            for key in keys:
                if key.key[0] in key_names:
                    raise ValueError(f"Duplicate sort key found: {key.key[0]}")
                key_names.append(key.key[0])

        return SortScheme(
            {
                "keys": keys,
                "name": name,
                "id": scheme_id,
                "nativeObject": native_object,
            }
        )

    def equivalent_to(
        self,
        other: SortScheme,
        check_identifiers: bool = False,
    ) -> bool:
        if other is None:
            return False
        if not isinstance(other, dict):
            return False
        if not isinstance(other, SortScheme):
            other = SortScheme(other)
        # If both have None keys, they are equivalent (for unsorted schemes)
        if self.keys is None and other.keys is None:
            return not check_identifiers or (
                self.name == other.name and self.id == other.id
            )
        # If only one has None keys, they are not equivalent
        if self.keys is None or other.keys is None:
            return False
        # Compare keys if both have them
        for i in range(len(self.keys)):
            if not self.keys[i].equivalent_to(other.keys[i]):
                return False
        return not check_identifiers or (
            self.name == other.name and self.id == other.id
        )

    @property
    def keys(self) -> Optional[SortKeyList]:
        val: List[SortKey] = self.get("keys")
        if val is not None and not isinstance(val, SortKeyList):
            self["keys"] = val = SortKeyList.of(val)
        return val

    @property
    def name(self) -> Optional[str]:
        return self.get("name")

    @property
    def id(self) -> Optional[str]:
        return self.get("id")

    @property
    def arrow(self) -> SortOptions:
        # TODO(pdames): Remove homogenous null ordering when supported by Arrow.
        if self.keys:
            if len(set([key.null_order for key in self.keys])) == 1:
                return SortOptions(
                    sort_keys=[pa_key for k in self.keys for pa_key in k.arrow],
                    null_placement=self.keys[0].null_order.value,
                )
            else:
                err_msg = "All arrow sort keys must use the same null order."
                raise ValueError(err_msg)
        return SortOptions()

    @property
    def native_object(self) -> Optional[Any]:
        return self.get("nativeObject")


UNSORTED_SCHEME = SortScheme.of(
    keys=None,
    name=UNSORTED_SCHEME_NAME,
    scheme_id=UNSORTED_SCHEME_ID,
)


class SortSchemeList(List[SortScheme]):
    @staticmethod
    def of(items: List[SortScheme]) -> SortSchemeList:
        typed_items = SortSchemeList()
        for item in items:
            if item is not None and not isinstance(item, SortScheme):
                item = SortScheme(item)
            typed_items.append(item)
        return typed_items

    def __getitem__(self, item):
        val = super().__getitem__(item)
        if val is not None and not isinstance(val, SortScheme):
            self[item] = val = SortScheme(val)
        return val

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]  # This triggers __getitem__ conversion



from typing import List, Optional
import logging
import pyarrow as pa
from deltacat import logs

from typing import Callable, Tuple
from deltacat.compute.compactor_v2.deletes.delete_file_envelope import (
    DeleteFileEnvelope,
)
from deltacat.compute.compactor_v2.deletes.delete_strategy import (
    DeleteStrategy,
)
import pyarrow.compute as pc
import numpy as np


logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class EqualityDeleteStrategy(DeleteStrategy):
    """
    A strategy for applying row-level deletes on tables during compaction based on equality conditions on one or more columns. It
    implements the "equality delete" approach, which marks a row as deleted by one or more column values like pk1=3 or col1="foo", col2="bar".

    Attributes:
        _name (str): The name of the delete strategy.

    Methods:
        name(self) -> str:
            Returns the name of the delete strategy.

        apply_deletes(self, table, delete_file_envelope, *args, **kwargs) -> Tuple[pa.Table, int]:
            Apply delete operations on the given table using the provided delete file envelope.

        apply_many_deletes(self, table, delete_file_envelopes, *args, **kwargs) -> Tuple[pa.Table, int]:
            Apply delete operations on the given table using all provided delete file envelopes.
    """

    _name = "EqualityDeleteStrategy"

    @property
    def name(self) -> str:
        """
        The name of the delete strategy.
        """
        return self._name

    def _drop_rows(
        self,
        table: pa.Table,
        delete_table: pa.Table,
        delete_column_names: List[str],
        equality_predicate_operation: Optional[Callable] = pa.compute.and_,
    ) -> Tuple[pa.Table, int]:
        """
        Drop rows from the given table based on the provided delete table and column names.

        Args:
            table (pa.Table): The input table to drop rows from.
            delete_table (pa.Table): The table containing the values to match for deletion.
            delete_column_names (List[str]): A list of column names to check for equality with the delete table.
            equality_predicate_operation (Optional[Callable], optional): The operation to combine equality predicates for multiple columns.
                Defaults to pa.compute.and_.

        Returns:
            Tuple[pa.Table, int]: A tuple containing the updated table after dropping rows,
                and the number of rows dropped.
        """
        if len(delete_column_names) < 1:
            return table, 0
        # all 1s -> all True so wont discard any from the curr_boolean_mask
        prev_boolean_mask = pa.array(np.ones(len(table), dtype=bool))
        # all 0s -> all False so if mask is never modified all rows will be kept
        curr_boolean_mask = pa.array(np.zeros(len(table), dtype=bool))
        for delete_column_name in delete_column_names:
            if delete_column_name not in table.column_names:
                logger.warning(
                    f"Column name {delete_column_name} not in table column names. Skipping dropping rows for this column."
                )
                continue
            curr_boolean_mask = pc.is_in(
                table[delete_column_name],
                value_set=delete_table[delete_column_name],
            )
            curr_boolean_mask = equality_predicate_operation(
                prev_boolean_mask, curr_boolean_mask
            )
            prev_boolean_mask = curr_boolean_mask
        number_of_rows_before_dropping = len(table)
        logger.debug(
            f"Number of table rows before dropping: {number_of_rows_before_dropping}."
        )
        table = table.filter(pc.invert(curr_boolean_mask))
        number_of_rows_after_dropping = len(table)
        logger.debug(
            f"Number of table rows after dropping: {number_of_rows_after_dropping}."
        )
        dropped_rows = number_of_rows_before_dropping - number_of_rows_after_dropping
        return table, dropped_rows

    def apply_deletes(
        self,
        table: Optional[pa.Table],
        delete_file_envelope: DeleteFileEnvelope,
        *args,
        **kwargs,
    ) -> Tuple[pa.Table, int]:
        """
        Apply delete operations on the given table using the provided delete file envelope.

        Args:
            table (Optional[pa.Table]): The pyarrow table to apply deletes on.
            delete_file_envelope (DeleteFileEnvelope): The delete file envelope containing delete parameters.

        Returns:
            Tuple[pa.Table, int]: A tuple containing the updated Arrow table after applying deletes,
                and the number of rows deleted.
        """
        if not table or not delete_file_envelope.table:
            logger.debug(
                f"No table passed or no delete file envelope delete table found. DeleteFileEnvelope: {delete_file_envelope}"
            )
            return table, 0
        delete_columns = delete_file_envelope.delete_columns
        delete_table = delete_file_envelope.table
        table, number_of_rows_dropped = self._drop_rows(
            table, delete_table, delete_columns
        )
        return table, number_of_rows_dropped

    def apply_many_deletes(
        self,
        table: Optional[pa.Table],
        delete_file_envelopes: List[DeleteFileEnvelope],
        *args,
        **kwargs,
    ) -> Tuple[pa.Table, int]:
        """
        Apply delete operations on the given table using all provided delete file envelopes.

        Args:
            table (Optional[pa.Table]): The pyarrow table to apply deletes on.
            delete_file_envelopes (List[DeleteFileEnvelope]): A list of delete file envelopes containing delete parameters.

        Returns:
            Tuple[pa.Table, int]: A tuple containing the updated pyarrow table after applying all deletes,
                and the total number of rows deleted.
        """
        # looking up table references are lighter than the actual table
        if not table:
            logger.debug("No table passed.")
            return table, 0
        total_dropped_rows = 0
        for delete_file_envelope in delete_file_envelopes:
            if delete_file_envelope.table_reference is None:
                continue
            table, number_of_rows_dropped = self.apply_deletes(
                table, delete_file_envelope
            )
            total_dropped_rows += number_of_rows_dropped
        return table, total_dropped_rows



import logging
from typing import Optional, Any, Set

from pyiceberg.catalog import Catalog
from pyiceberg.table import Table
import deltacat.logs as logs

from deltacat.storage.model.scan.push_down import Pushdown, PartitionFilter
from deltacat.storage.model.scan.scan_plan import ScanPlan
from deltacat.storage.model.scan.scan_task import FileScanTask, DataFile
from deltacat.storage.util.scan_planner import ScanPlanner
from deltacat.experimental.storage.iceberg.impl import _try_load_iceberg_table
from deltacat.experimental.storage.iceberg.visitor import IcebergExpressionVisitor

# Initialize DeltaCAT logger
logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class IcebergScanPlanner(ScanPlanner):
    def __init__(self, catalog: Catalog):
        self.catalog = catalog
        self.expression_visitor = IcebergExpressionVisitor()

    @classmethod
    def _collect_filter_fields(cls, expr: Any) -> Set[str]:
        """
        Collects all field names referenced in the filter expression.

        Args:
            expr: The expression to analyze

        Returns:
            Set of field names referenced in the expression
        """
        fields = set()
        if hasattr(expr, "field"):
            fields.add(expr.field)
        if hasattr(expr, "left"):
            fields.update(cls._collect_filter_fields(expr.left))
        if hasattr(expr, "right"):
            fields.update(cls._collect_filter_fields(expr.right))
        if hasattr(expr, "expr"):
            fields.update(cls._collect_filter_fields(expr.expr))
        if hasattr(expr, "values"):
            for value in expr.values:
                fields.update(cls._collect_filter_fields(value))
        return fields

    def create_scan_plan(
        self,
        table_name: str,
        namespace: Optional[str] = None,
        pushdown: Optional[Pushdown] = None,
    ) -> ScanPlan:
        iceberg_table = _try_load_iceberg_table(
            self.catalog, namespace=namespace, table_name=table_name
        )

        # TODO: implement row, column predicate pushdown to Iceberg

        # Get the partition spec
        partition_spec = iceberg_table.spec()

        # Check if the table is partitioned
        is_partitioned = len(partition_spec.fields) > 0

        scan = iceberg_table.scan()
        if is_partitioned:
            if pushdown and pushdown.partition_filter:
                filter_fields = self._collect_filter_fields(pushdown.partition_filter)
                logger.info(
                    f"Pushdown partition filter is enabled, converting to Iceberg. Fields discovered in filter: {', '.join(sorted(filter_fields))}"
                )
                # Handle partition filter if present, DeltaCAT only supports partition-level filters right now
                iceberg_expression = self._convert_partition_filter(
                    iceberg_table, pushdown.partition_filter
                )
                scan = scan.filter(iceberg_expression)

        file_scan_tasks = []
        for scan_task in scan.plan_files():
            file_scan_tasks.append(FileScanTask([DataFile(scan_task.file.file_path)]))
        return ScanPlan(file_scan_tasks)

    @classmethod
    def _validate_partition_references(
        cls, expr: Any, partition_cols: Set[str]
    ) -> None:
        """
        Validates that the expression only references partition columns.

        Args:
            expr: The expression to validate
            partition_cols: Set of valid partition column names

        Raises:
            ValueError: If the expression references a non-partition column
        """
        if hasattr(expr, "field"):  # Reference type expression
            if expr.field not in partition_cols:
                raise ValueError(
                    f"Filter references non-partition column: {expr.field}. "
                    f"Partition columns are: {partition_cols}"
                )
        # Recursively validate nested expressions
        if hasattr(expr, "left"):
            cls._validate_partition_references(expr.left, partition_cols)
        if hasattr(expr, "right"):
            cls._validate_partition_references(expr.right, partition_cols)
        if hasattr(expr, "expr"):
            cls._validate_partition_references(expr.expr, partition_cols)
        if hasattr(expr, "values"):
            for value in expr.values:
                cls._validate_partition_references(value, partition_cols)

    def _convert_partition_filter(
        self, table: Table, partition_filter: PartitionFilter
    ):
        """
        Convert DeltaCAT partition filter to PyIceberg expression,
        validating that only partition columns are referenced.
        """
        partition_cols = set(field.name for field in table.spec().fields)

        # Validate before converting
        self._validate_partition_references(partition_filter, partition_cols)

        # Convert to PyIceberg expression
        return self.expression_visitor.visit(partition_filter)



import time
import os
import posixpath
import pyarrow.fs
from pyarrow.fs import FileSelector, FileType
from itertools import chain
from deltacat.storage.model.transaction import Transaction
from deltacat.utils.filesystem import resolve_path_and_filesystem
from deltacat.constants import (
    TXN_DIR_NAME,
    RUNNING_TXN_DIR_NAME,
    FAILED_TXN_DIR_NAME,
    TXN_PART_SEPARATOR,
)
from deltacat.storage.model.types import TransactionState
import logging
from deltacat import logs

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def brute_force_search_matching_metafiles(
    dirty_files_names, filesystem: pyarrow.fs.FileSystem, catalog_root
):
    txn_dir_name = TXN_DIR_NAME
    # collect transaction ids of the files
    transaction_ids = []
    for dirty_file in dirty_files_names:
        parts = dirty_file.split(TXN_PART_SEPARATOR)
        if len(parts) < 2:
            continue
        transaction_ids.append(parts[1])

    def recursive_search(path):
        try:
            selector = FileSelector(path, recursive=False)
            entries = filesystem.get_file_info(selector)
        except Exception as e:
            logger.error(f"Error listing directory '{path}': {e}")
            return

        for entry in entries:
            base_name = posixpath.basename(entry.path)
            if entry.type == FileType.File:
                for transaction_id in transaction_ids:
                    # Look for transaction_id in the filename
                    if transaction_id in base_name:
                        try:
                            filesystem.delete_file(entry.path)
                            logger.debug(f"Deleted file: {entry.path}")
                        except Exception as e:
                            logger.error(f"Error deleting file '{entry.path}': {e}")

            elif entry.type == FileType.Directory:
                # Skip directories that match txn_dir_name
                if posixpath.basename(entry.path) == txn_dir_name:
                    logger.debug(f"Skipping directory: {entry.path}")
                    continue
                recursive_search(entry.path)

    # Start recursive search from the catalog root
    recursive_search(catalog_root)

    # renaming to successful completion
    for dirty_file in dirty_files_names:
        failed_txn_log_dir = posixpath.join(
            catalog_root, TXN_DIR_NAME, FAILED_TXN_DIR_NAME
        )
        old_log_path = posixpath.join(failed_txn_log_dir, dirty_file)

        # new_filename = dirty_file.replace(TIMEOUT_TXN, SUCCESSFULLY_CLEANED)
        new_log_path = posixpath.join(failed_txn_log_dir, dirty_file)
        try:
            filesystem.move(old_log_path, new_log_path)
            logger.debug(f"Renamed file from {old_log_path} to {new_log_path}")
        except Exception as e:
            logger.error(f"Error renaming file '{old_log_path}': {e}")


def janitor_delete_timed_out_transaction(catalog_root: str) -> None:
    """
    Traverse the running transactions directory and move transactions that have been
    running longer than the threshold into the failed transactions directory.
    """
    catalog_root_normalized, filesystem = resolve_path_and_filesystem(catalog_root)

    txn_log_dir = posixpath.join(catalog_root_normalized, TXN_DIR_NAME)
    running_txn_log_dir = posixpath.join(txn_log_dir, RUNNING_TXN_DIR_NAME)
    failed_txn_log_dir = posixpath.join(txn_log_dir, FAILED_TXN_DIR_NAME)

    dirty_files = []

    running_txn_file_selector = FileSelector(running_txn_log_dir, recursive=False)
    running_txn_info_list = filesystem.get_file_info(running_txn_file_selector)

    for running_txn_info in running_txn_info_list:
        try:
            filename = posixpath.basename(running_txn_info.path)
            parts = filename.split(TXN_PART_SEPARATOR)
            end_time_str = parts[-1]
            end_time = float(end_time_str)
            current_time = time.time_ns()
            if end_time <= current_time:
                src_path = running_txn_info.path
                new_filename = f"{filename}"
                dest_path = posixpath.join(failed_txn_log_dir, new_filename)

                # Move the file using copy and delete
                with filesystem.open_input_file(src_path) as src_file:
                    contents = src_file.read()

                with filesystem.open_output_stream(dest_path) as dest_file:
                    dest_file.write(contents)
                filesystem.delete_file(src_path)

                dirty_files.append(new_filename)

        except Exception as e:
            logger.error(
                f"Error cleaning failed transaction '{running_txn_info.path}': {e}"
            )

    # Pass catalog_root to the brute force search so it searches from the right place
    brute_force_search_matching_metafiles(
        dirty_files, filesystem, catalog_root_normalized
    )


def janitor_remove_files_in_failed(
    catalog_root: str, filesystem: pyarrow.fs.FileSystem = None
) -> None:
    """
    Cleans up metafiles and locator files associated with failed transactions.
    """
    if filesystem is None:
        catalog_root_normalized, filesystem = resolve_path_and_filesystem(catalog_root)
    else:
        catalog_root_normalized, filesystem = resolve_path_and_filesystem(
            catalog_root, filesystem
        )

    txn_log_dir = posixpath.join(catalog_root_normalized, TXN_DIR_NAME)
    failed_txn_log_dir = posixpath.join(txn_log_dir, FAILED_TXN_DIR_NAME)
    running_txn_log_dir = posixpath.join(txn_log_dir, RUNNING_TXN_DIR_NAME)
    filesystem.create_dir(failed_txn_log_dir, recursive=True)

    failed_txn_file_selector = FileSelector(failed_txn_log_dir, recursive=False)
    failed_txn_info_list = filesystem.get_file_info(failed_txn_file_selector)

    for failed_txn_info in failed_txn_info_list:
        try:
            txn = Transaction.read(failed_txn_info.path, filesystem)
            failed_txn_basename = posixpath.basename(failed_txn_info.path)
            should_process = True
            try:
                if txn.state(catalog_root_normalized) == TransactionState.PURGED:
                    should_process = False
            except Exception:
                logger.error("Could not check attribute")
            if should_process:
                # Process if the file is marked as currently cleaning.
                txnid = txn.id

                if txn.state(catalog_root_normalized) == TransactionState.FAILED:

                    txnid = txn.id

                    operations = txn["operations"]
                    known_write_paths = chain.from_iterable(
                        (op["metafile_write_paths"] + op["locator_write_paths"])
                        for op in operations
                    )

                    for write_path in known_write_paths:
                        full_path = posixpath.join(catalog_root_normalized, write_path)
                        try:
                            filesystem.delete_file(full_path)
                        except Exception as e:
                            logger.error(f"Failed to delete file '{full_path}': {e}")

                    new_filename = f"{txnid}"

                    new_failed_txn_log_file_path = posixpath.join(
                        failed_txn_log_dir, new_filename
                    )
                    running_txn_log_path = posixpath.join(
                        running_txn_log_dir, new_filename
                    )

                    os.delete(running_txn_log_path)

                    os.rename(failed_txn_info.path, new_failed_txn_log_file_path)
                    logger.debug(
                        f"Cleaned up failed transaction: {failed_txn_basename}"
                    )

        except Exception as e:
            logger.error(
                f"Could not read transaction '{failed_txn_info.path}', skipping: {e}"
            )


def janitor_job(catalog_root_dir: str) -> None:
    janitor_delete_timed_out_transaction(catalog_root_dir)
    janitor_remove_files_in_failed(catalog_root_dir)
